/*
 * code.c --- routines for loading the firmware code
 * 
 * Written by Theodore Ts'o, 1999
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "loadrm2.h"

void free_code(code_t c)
{
	if (c) {
		if (c->name)
			free(c->name);
		if (c->data)
			free(c->data);
		free(c);
	}
}


int read_code(char *name, code_t *ret_code)
{
	code_t	c = 0;
	int	errcode, fd, len, left, i;
	unsigned char	*cp;
	unsigned long	xsum;
	struct	stat	st;

	if (stat(name, &st) < 0)
		return errno;

	if ((fd = open(name, O_RDONLY)) < 0)
		return errno;

	errcode = ENOMEM;
	c = malloc(sizeof(struct code));
	if (!c)
		goto errout;
	memset(c, 0, sizeof(struct code));
	c->name = malloc(strlen(name)+1);
	if (!c->name)
		goto errout;
	c->data = malloc(st.st_size);
	if (!c->data)
		goto errout;
	c->length = st.st_size;

	strcpy(c->name, name);

	for (left = c->length, cp = c->data; left > 0;) {
		len = read(fd, cp, left);
		if (len < 0) {
			if ((errno == EAGAIN) || (errno == EINTR))
				continue;
			errcode = errno;
			goto errout;
		}
		left -= len;
		cp += len;
	}
	/* calculate the checksum */
	for (i=0, xsum = 0x8000; i < c->length; i++) {
		if ((i == 0xFFBE) || (i == 0xFFBF))
			continue;
		xsum += c->data[i];
	}
	c->checksum = xsum & 0xFFFF;
	*ret_code = c;
	return 0;

errout:
	free_code(c);
	return errcode;
}



