/*
 * rcktldr.c --- Rocket Modem II microcode loader
 *
 * Written by Theodore Ts'o, 1999
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include "loadrm2.h"
#include "rocket.h"

code_t loader_code, firmware_code;
int	debug_flag, ati_flag, progress_flag;

static void usage()
{
    printf("Your command-line syntax....\n");
    exit(1);
}

//  Opens all the /dev/ttyRx ports
void get_all_ports(mux m)
{
    struct rocket_ports rocketInfo;
    int	board, line;
    char	buf[32];
    int	fd, ret;
    int    	head_msg=1;
    int	nl_count = 6;

    //  Open ttyR0, use it to get the config.  This tells
    //  how many ttyRx ports there are and what type they are
    fd = open("/dev/ttyR0", O_RDONLY);
    if (fd < 0) {
	perror("/dev/ttyR0");
	exit(1);
    }
    if (ioctl(fd, RCKP_GET_PORTS, &rocketInfo) < 0) {
	perror("Cannot get rocket info structure");
	close(fd);
	return;
    }

    //  Walk through all boards, opening the ports
    for (board = 0; board < 4; board++) {
	if (rocketInfo.rocketModel[board].loadrm2) {
	    for (line = rocketInfo.rocketModel[board].startingPortNumber;
		 line < rocketInfo.rocketModel[board].startingPortNumber +
		     rocketInfo.rocketModel[board].numPorts;
		 line++) {
		if (head_msg) {
		    printf("Initializing RocketModemII's:");
		    head_msg = 0;
		}
		if (nl_count--)
		    fputs(" ", stdout);
		else {
		    fputs("\n\t", stdout);
		    nl_count = 8;
		}
		sprintf(buf, "/dev/ttyR%d", line);
		fputs(buf+5, stdout);
		ret = port_open(m, buf);
		if (ret) {
		    fputs("\n\n", stdout);
		    fprintf(stderr,
			    "Couldn't open %s: %s\n",
			    buf, strerror(ret));
		    exit(1);
		}
	    }
	}
    }

    if (head_msg == 0)
	printf("\n");
}


int main(int argc, char **argv)
{
    int	c, ret;
    mux	m;
    const static char *errmsg =
	"Couldn't read Rocketmodem II %s file (%d)\n";
	
    while ((c = getopt(argc, argv, "dip")) != EOF) {
	switch (c) {
	case 'd':
	    debug_flag++;
	    break;
	case 'i':
	    ati_flag++;
	    break;
	case 'p':
	    progress_flag++;
	    break;
	default:
	    usage();
			
	}
    }
	
    ret = read_code(PATH_CTMLDR, &loader_code);
    if (ret) {
	printf(errmsg, "loader", ret);
	exit(1);
    }
    if (debug_flag)
	printf("Loader size=%d, checksum=0x%lx\n",
	       (int)loader_code->length, loader_code->checksum);
    ret = read_code(PATH_CTMFW, &firmware_code);
    if (ret) {
	printf(errmsg, "firmware", ret);
	exit(1);
    }
    if (debug_flag)
	printf("Firmware size=%d, checksum=0x%lx\n",
	       (int)firmware_code->length, firmware_code->checksum);

    mux_init(&m);

    if (argv[optind]) {
	while (argv[optind]) {
	    ret = port_open(m, argv[optind]);
	    if (ret) {
		fprintf(stderr, "Couldn't open %s: %s",
			argv[optind], strerror(ret));
		exit(1);
	    }
	    optind++;
	}
    } else {
	get_all_ports(m);
    }
    if (num_ports == 0)
	return 0;

    mux_loop(m);
    if (num_ports)
	progress_end();

    return 0;
}

      
