#!/usr/bin/perl
# This little hack program will take one argument in the form of a kernel version number (ie 2.6.35) and 
#  strip out all code from a .c file that is not needed by the specified kernel number.
# The purpose of this is to prepare the driver for inclusion in the upstream kernel source

my $file = 'rocket'; # without the .c extension

my $debug = 0;
my $major, $minor, $patchlvl;

if ($ARGV[0] !~ /\d+\.\d+\.\d+/) {
  print "Usage: $0 <kernel version number ie 2.4.25>\n";
  print "   $0 2.4.27\n\n";
  exit;
}

($major, $minor, $patchlvl) = split (/\./, $ARGV[0]);
print "DEBUG: argument: major: $major minor: $minor patch: $patchlvl\n" if $debug;
my $outfile = $file."-$major.$minor.$patchlvl.c";


# read in the whole $file

open IN, "<$file.c";
@origdriver = (<IN>);
close IN;

open OUT, ">$outfile";
$toggle = 1;
$nested = 0;
$versionif = 0;
$linenum = 0;
foreach $line (@origdriver) {
  # search for lines where LINUX_VERSION_CODE is used
  if ($line =~ /^#if LINUX_VERSION_CODE/i) {
    ($tempversion) = ( $line =~ /VERSION_CODE\((\d+\,\d+\,\d+)\)/i);
    ($tempmajor, $tempminor, $temppatch) = split /,/, $tempversion;
    ($tempoperator) = ( $line =~ /LINUX_VERSION_CODE\s*(\S*)\s*VERSION_CODE/i);
    $nested++;
    print "DEBUG: $nested: line: $linenum major: $tempmajor minor: $tempminor patch: $temppatch operator: $tempoperator\n" if $debug;
    if ($tempoperator eq "<") {
      if ($major < $tempmajor) {
        print "DEBUG: MAJOR TOGGLE ON!\n" if $debug;
        $toggle = 1;
      }
      elsif ($major == $tempmajor) {
        if ($minor < $tempminor) {
          print "DEBUG: MINOR TOGGLE ON!\n" if $debug;
          $toggle = 1;
        }
        elsif ($minor == $tempminor) {
          if ($patchlvl < $temppatch) {
            print "DEBUG: PATCH TOGGLE ON!\n" if $debug;
            $toggle = 1;
          }
          else {
            print "DEBUG: PATCH TOGGLE OFF!\n" if $debug;
            $toggle = 0;
          }
        }
        else {
          print "DEBUG: MINOR TOGGLE OFF!\n" if $debug;
          $toggle = 0;
        }
      }
      else {
        print "DEBUG: MAJOR TOGGLE OFF!\n" if $debug;
        $toggle = 0;
      }
    }
    elsif ($tempoperator eq ">") {
      if ($major > $tempmajor) {
        print "DEBUG: MAJOR TOGGLE ON!\n" if $debug;
        $toggle = 1;
      }
      elsif ($major == $tempmajor) {
        if ($minor > $tempminor) {
          print "DEBUG: MINOR TOGGLE ON!\n" if $debug;
          $toggle = 1;
        }
        elsif ($minor == $tempminor) {
          if ($patchlvl > $temppatch) {
            print "DEBUG: PATCH TOGGLE ON!\n" if $debug;
            $toggle = 1;
          }
          else {
            print "DEBUG: PATCH TOGGLE OFF!\n" if $debug;
            $toggle = 0;
          }
        }
        else {
          print "DEBUG: MINOR TOGGLE OFF!\n" if $debug;
          $toggle = 0;
        }
      }
      else {
        print "DEBUG: MAJOR TOGGLE OFF!\n" if $debug;
        $toggle = 0;
      }
    }
    #push @openif, $linenum;
    $versionif[$nested] = 1;
    $nonversionif = 0;
  }
  elsif ($line =~ /^#if/) {
    $nested++;
    #print "DEBUG: $nested: line: $linenum if thats not important\n" if $debug;
    $versionif[$nested] = 0;
    if ($toggle == 1) {
      print "DEBUG: IF PRINTED!\n" if $debug;
      print OUT $line;
    }
  }
  elsif ($line =~ /^#else/) {
    if ($versionif[$nested]) {
      print "DEBUG: $nested: line: $linenum else\n" if $debug;
      if ($toggle == 0) {
        $toggle = 1;
        print "DEBUG: ELSE TOGGLE ON!\n" if $debug;
      }
      elsif ($toggle == 1) {
        $toggle = 0;
        print "DEBUG: ELSE TOGGLE OFF!\n" if $debug;
      }
    }
    elsif ($toggle == 1) {
      print "DEBUG: ELSE PRINTED!\n" if $debug;
      print OUT $line;
    }
  }
  elsif ($line =~ /^#endif/) {
    if ($versionif[$nested]) {
      print "DEBUG: $nested: line: $linenum endif\n" if $debug;
    }
    elsif ($toggle == 1) {
      print "DEBUG: ENDIF PRINTED!\n" if $debug;
      print OUT $line;
    }
    $versionif[$nested] = 0;
    $nested--;
    if ($versionif[$nested] == 0) { $toggle = 1; }
  }
  elsif ($toggle == 1) {
    if (! $printedcode) {
      print "DEBUG: CODE INCLUDED!\n" if $debug;
      $printedcode = 1;
    }
    print OUT $line;
  }
  else { $printedcode = 0; }
  $linenum++;
}

close OUT;

if ($nested != 0) {
  print "There are 'if' statements that have no 'endif'!\n";
}

print "Wrote out to '$outfile'\n";
