/* setrocket.c --- get/set Rocket flags */
//krw - 2006-05-04 - added stdlib.h declare to eliminate warnings about exit(n) incompatible implicit declaration of built-in function exit
#include <stdlib.h>	
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>

#define MKDEV(a,b) ((int)((((a) & 0xff) << 8) | ((b) & 0xff)))

#include "rocket.h"
#include "socketmodem.h"

#define VERSION_STR "1.5"

char *progname;

int	verbosity = 1;		/* 1 = normal, 0=boot-time, 2=everything */
int	verbose_flag = 0;	/* print results after setting a port */
int	quiet_flag = 0;
int	old_device_name = 0;

#define CMD_FLAG	1
#define CMD_DELAY	2
#define CMD_WAIT	3
#define CMD_COUNTRY	4

#define FLAG_CAN_INVERT	0x0001
#define FLAG_NEED_ARG	0x0002

struct flag_type_table {
    int	cmd;
    char	*name;
    int	bits;
    int	mask;
    int	level;
    int	flags;
} flag_type_tbl[] = {
    CMD_FLAG,	"rtsToggle",	ROCKET_RTS_TOGGLE, ROCKET_RTS_TOGGLE, 0, FLAG_CAN_INVERT,
    CMD_FLAG,	"spd_normal",	0,		ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_hi",	ROCKET_SPD_HI, 	ROCKET_SPD_MASK,        0, 0,
    CMD_FLAG,	"spd_vhi",	ROCKET_SPD_VHI,	ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_shi",	ROCKET_SPD_SHI,	ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_warp",	ROCKET_SPD_WARP, ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_warp13",	ROCKET_SPD_WARP13, ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,       "rs232",        ROCKET_MODE_RS232, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,       "rs422",        ROCKET_MODE_RS422, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,       "rs485",        ROCKET_MODE_RS485, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG, 	"SAK", 		ROCKET_SAK, 	ROCKET_SAK, 	0, FLAG_CAN_INVERT,
    CMD_FLAG,	"hup_notify",	ROCKET_HUP_NOTIFY, ROCKET_HUP_NOTIFY, 0, FLAG_CAN_INVERT,
    CMD_FLAG,	"split_termios", ROCKET_SPLIT_TERMIOS, ROCKET_SPLIT_TERMIOS, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"session_lockout", ROCKET_SESSION_LOCKOUT, ROCKET_SESSION_LOCKOUT, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"pgrp_lockout", ROCKET_PGRP_LOCKOUT, ROCKET_PGRP_LOCKOUT, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"callout_nohup", ROCKET_CALLOUT_NOHUP, ROCKET_CALLOUT_NOHUP, 2, FLAG_CAN_INVERT,

    CMD_DELAY,	"close_delay",	0,		0,		0, FLAG_NEED_ARG,
    CMD_WAIT,	"closing_wait",	0,		0,		0, FLAG_NEED_ARG,
    CMD_COUNTRY,	"country",	0,		0,		0, FLAG_NEED_ARG,
    0,		0,		0,		0,		0, 0,
};
	
struct country_code_table {
    int	code;
    char	*name;
} country_code_tbl[] = {
    mcAustralia,	"Australia",
    mcAustria,	"Austria",
    mcBelgium,	"Belgium",
    mcBulgaria,	"Bulgaria",
    mcCanada,	"Canada",
    mcChina,	"China",
    mcCzechRep,	"CzechRep",
    mcDenmark,	"Denmark",
    mcFinland,	"Finland",
    mcFrance,	"France",
    mcGermany,	"Germany",
    mcGreece,	"Greece",
    mcHongKong,	"HongKong",
    mcHungary,	"Hungary",
    mcIndia,	"India",
    mcIreland,	"Ireland",
    mcIsrael,	"Israel",
    mcItaly,	"Italy",
    mcJapan,	"Japan",
    mcKorea,	"Korea",
    mcLuxembourg,	"Luxembourg",
    mcMexico,	"Mexico",
    mcNA,		"NA",
    mcNetherlands,	"Netherlands",
    mcNewZealand,	"NewZealand",
    mcNorway,	"Norway",
    mcPhilippines,	"Philippines",
    mcPoland,	"Poland",
    mcPortugal,	"Portugal",
    mcRussia,	"Russia",
    mcSingapore,	"Singapore",
    mcSlovacRep,	"SlovacRep",
    mcSpain,	"Spain",
    mcSweden,	"Sweden",
    mcSwitzerland,	"Switzerland",
    mcTaiwan,	"Taiwan",
    mcUK,           "United Kingdom",
    mcUSA,		"USA",
    0, 0, };
 	
void signal_alarm(int signum)
{
    fprintf(stderr, "Timeout communicating with modem\n");
    exit(1);
}

void set_timeout(int timeout)
{
    struct sigaction	sa;

    if (!timeout) {
	alarm(0);
	return;
    }			
    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = signal_alarm;
    if (sigaction(SIGALRM, &sa, NULL) < 0) {
	fprintf(stderr, "Error setting alarm signal: %s\n",
		strerror(errno));
	exit(1);
    }
    alarm(timeout);
}

int atonum(char *s)
{
    int n;

    while (*s == ' ')
	s++;
    if (strncmp(s, "0x", 2) == 0 || strncmp(s, "0X", 2) == 0)
	sscanf(s + 2, "%x", &n);
    else if (s[0] == '0' && s[1])
	sscanf(s + 1, "%o", &n);
    else
	sscanf(s, "%d", &n);
    return n;
}

void print_flags(struct rocket_config *serinfo,
		 char *prefix, char *postfix)
{
    struct	flag_type_table	*p;
    int	flags;
    int	first = 1;

    flags = serinfo->flags;
	
    for (p = flag_type_tbl; p->name; p++) {
	if (p->cmd != CMD_FLAG)
	    continue;
	if (verbosity < p->level)
	    continue;
	if ((flags & p->mask) == p->bits) {
	    if (first) {
		printf("%s", prefix);
		first = 0;
	    } else
		printf(" ");
	    printf("%s", p->name);
	}
    }
	
    if (!first)
	printf("%s", postfix);
}

/*
 * Set the termios to something sane
 */
int initialize_termios(int fd)
{
    struct termios	t;

    if (tcgetattr(fd, &t) < 0)
	return errno;

    cfsetospeed(&t, B9600);
    cfsetispeed(&t, B9600);
    t.c_iflag &= ~(ISTRIP|INLCR|ICRNL|IGNCR);
    t.c_iflag |= ICRNL;
    t.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET|OFILL);
    t.c_cflag |= CREAD|CLOCAL;
    t.c_lflag &= ~(ISIG|ICANON|ECHO|IEXTEN);

    if (tcsetattr(fd, TCSANOW, &t) < 0)
	return errno;

    return 0;
}

void set_country(int fd, char *country)
{
    struct country_code_table *ccp;
    struct termios new, old;
    int code;
    int	newfd;
    FILE	*f;
    char	buf[256];

    if (country[0] == '?') {
	for (ccp = country_code_tbl; ccp->name; ccp++) {
	    printf("%s\n",ccp->name);
	}
	exit(1);
    }

    for (ccp = country_code_tbl; ccp->name; ccp++) {
	if (!strcasecmp(ccp->name, country))
	    break;
    }
    if (!ccp->name) {
	fprintf(stderr, "Invalid country name: %s\n", country);
	exit(1);
    }
    code = ccp->code;

    if (tcgetattr(fd, &old) < 0) {
	perror("tcgetattr");
	exit(1);
    }

    new = old;
    new.c_iflag &= ~(ISTRIP|INLCR|ICRNL|IGNCR);
    new.c_iflag |= ICRNL;
    new.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET|OFILL);
    new.c_cflag |= CREAD|CLOCAL;
    new.c_lflag &= ~(ISIG|ICANON|ECHO|IEXTEN);
	
    if (tcsetattr(fd, TCSANOW, &new) < 0) {
	perror("tcsetattr");
	exit(1);
    }

    newfd = dup(fd);
    if (newfd < 0) {
	perror("dup");
	exit(1);
    }
    f = fdopen(newfd, "w+");
    if (!f) {
	perror("fdopen");
	exit(1);
    }
    setlinebuf(f);

    fprintf(f, "\r\nAT*NC%02dZ\r\n", code);
    fflush(f);
    fseek(f, 0L, SEEK_CUR);
    set_timeout(30);
    while(1) {
	memset(buf, 0, sizeof(buf));
	if (fgets(buf, sizeof(buf), f) == NULL) {
	    fprintf(stderr, "Couldn't read results "
		    "from serial port.\n");
	    exit(1);
	}
	if (!strncasecmp(buf, "ok", 2))
	    break;
	if (!strncasecmp(buf, "error", 5)) {
	    fprintf(stderr, "Modem reported error "
		    "setting country code %d\n", code);
	    break;
	}
    }
    set_timeout(0);
    fclose(f);

    if (tcsetattr(fd, TCSANOW, &old) < 0) {
	perror("tcsetattr");
	exit(1);
    }

}


void get_rocket(char *device)
{
    struct rocket_config config;
    int	fd;
    char	buf1[40];

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	perror(device);
	return;
    }
    if (ioctl(fd, RCKP_GET_CONFIG, &config) < 0) {
	perror("Cannot get rocket config structure");
	close(fd);
	return;
    }
    if (verbosity==2) {
	printf("%s, Line %d, Base port: 0x%.4x\n",
	       device, config.line, config.port);
	if (config.closing_wait == ROCKET_CLOSING_WAIT_INF)
	    strcpy(buf1, "infinte");
	else if (config.closing_wait == ROCKET_CLOSING_WAIT_NONE)
	    strcpy(buf1, "none");
	else
	    sprintf(buf1, "%d", config.closing_wait);
	printf("\tclose_delay: %d, closing_wait: %s\n",
	       config.close_delay, buf1);
	print_flags(&config, "\tFlags: ", "");
	printf("\n\n");
    } else {
	printf("%s, ", device);
	print_flags(&config, "Flags: ", "");
	printf("\n");
    }
    close(fd);
}

void set_rocket(char *device, char ** arg)
{
    struct rocket_config old_config, new_config;
    struct	flag_type_table	*p;
    int	fd;
    int	do_invert = 0;
    char	*word;
	

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	if (verbosity==0 && errno==ENOENT)
	    exit(201);
	perror(device);
	exit(201);
    }
    if (fcntl(fd, F_SETFL, 0) < 0) {
	perror("fcntl");
	exit(1);
    }
    if (ioctl(fd, RCKP_GET_CONFIG, &old_config) < 0) {
	perror("Cannot get rocket info");
/*		exit(1); */
    }
    new_config = old_config;
    while (*arg) {
	do_invert = 0;
	word = *arg++;
	if (*word == '^') {
	    do_invert++;
	    word++;
	}
	for (p = flag_type_tbl; p->name; p++) {
	    if (!strcasecmp(p->name, word))
		break;
	}
	if (!p->name) {
	    fprintf(stderr, "Invalid flag: %s\n", word);
	    exit(1);
	}
	if (do_invert && !(p->flags & FLAG_CAN_INVERT)) {
	    fprintf(stderr, "This flag can not be inverted: %s\n", word);
	    exit(1);
	}
	if ((p->flags & FLAG_NEED_ARG) && !*arg) {
	    fprintf(stderr, "Missing argument for %s\n", word);
	    exit(1);
	}
	switch (p->cmd) {
	case CMD_FLAG:
	    new_config.flags &= ~p->mask;
	    if (!do_invert)
		new_config.flags |= p->bits;
	    break;
	case CMD_DELAY:
	    new_config.close_delay = atonum(*arg++);
	    break;
	case CMD_WAIT:
	    if (!strcasecmp(*arg, "infinite"))
		new_config.closing_wait = ROCKET_CLOSING_WAIT_INF;
	    else if (!strcasecmp(*arg, "none"))
		new_config.closing_wait = ROCKET_CLOSING_WAIT_NONE;
	    else
		new_config.closing_wait = atonum(*arg);
	    arg++;
	    break;
	case CMD_COUNTRY:
	    set_country(fd, *arg++);
	    break;
	default:
	    fprintf(stderr, "Internal error: unhandled cmd #%d\n", p->cmd);
	    exit(1);
	}
    }
    if (ioctl(fd, RCKP_SET_CONFIG, &new_config) < 0) {
	perror("Cannot set rocket info");
	exit(1);
    }
    close(fd);
    if (verbose_flag)
	get_rocket(device);
}

void make_device(char *pathname, int mode, int major, int minor,
		 gid_t tty_group)
{
    struct stat statbuf;
    dev_t	device;

    device = MKDEV(major, minor);
    if (lstat(pathname, &statbuf) == 0) {
	if (device == statbuf.st_rdev)
	    return;
	unlink(pathname);
    }
    if (mknod(pathname, mode, device) < 0) {
	perror(pathname);
    }
    chown(pathname, 0, tty_group);
}

void mkdev_rocket(char *device)
{
    struct rocket_version rocketVer;
    struct rocket_ports rocketInfo;
    int	fd, board, port, line;
    gid_t tty_group = 0;
    struct group *grp;
    char pathname[64];
    mode_t mode;

    grp = getgrnam("tty");
    if (grp)
	tty_group = grp->gr_gid;

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	perror(device);
	return;
    }
    if (ioctl(fd, RCKP_GET_VERSION, &rocketVer) < 0) {
        perror("Cannot get rocket driver version");
        return;
    }
    if (ioctl(fd, RCKP_GET_PORTS, &rocketInfo) < 0) {
	perror("Cannot get rocket ports structure");
	close(fd);
	return;
    }
    close(fd);
    mode = S_IFCHR | 0620;
    umask(0);
    fflush(stdout);

    printf("RocketPort device driver module, version %s, %s\n",
	   rocketVer.rocket_version, rocketVer.rocket_date);

    // Walk through the model info struct a board at a time, creating
    // the /dev/ttyRx and /dev/cuRx devices
    for (board = 0; board < 4; board++) {
	if (!rocketInfo.rocketModel[board].model)
	    break;

	printf("Installing %s, creating /dev/ttyR%d - %d\n", 
	       rocketInfo.rocketModel[board].modelString,
	       rocketInfo.rocketModel[board].startingPortNumber,
	       rocketInfo.rocketModel[board].startingPortNumber + 
	       rocketInfo.rocketModel[board].numPorts - 1);
	for (line = rocketInfo.rocketModel[board].startingPortNumber;
	     line < rocketInfo.rocketModel[board].startingPortNumber + 
		 rocketInfo.rocketModel[board].numPorts; line++) {

	    //  Create the /dev/ttyR0 device
	    sprintf(pathname, "/dev/ttyR%d", line);
//	    printf("Creating %s \n", pathname);
	    make_device(pathname, mode, rocketInfo.tty_major,
			line, tty_group);
    	}
    }
}

void usage()
{
    fprintf(stderr, "setrocket Version %s\n\n", VERSION_STR);
    fprintf(stderr,
	    "usage: %s rocket-device [cmd1 [arg]] ... \n\n", progname);
    fprintf(stderr, "Available commands: (* = Takes an argument)\n");
    fprintf(stderr, "\t\t(^ = can be preceded by a '^' to turn off the option)\n");
    fprintf(stderr, "\t* close_delay\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that DTR should be kept low\n");
    fprintf(stderr, "\t\t\t\twhile being closed\n");
    fprintf(stderr, "\t* closing_wait\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that the rocket port should wait for\n");
    fprintf(stderr, "\t\t\t\tdata to be drained while being closed\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t^ rtsToggle\tEnable the RTS toggle flag on the port\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t  rs232\t\tSet port interface hardware to RS-232\n");
    fprintf(stderr, "\t  rs422\t\tSet port interface hardware to RS-422\n");
    fprintf(stderr, "\t  rs485\t\tSet port interface hardware to RS-485\n");
    fprintf(stderr, "\n");
#if 0
    fprintf(stderr, "\t^ sak\t\tset the break key as the Secure Attention Key\n");
#endif
    fprintf(stderr, "\t^ session_lockout Lock out callout port across different sessions\n");
    fprintf(stderr, "\t^ pgrp_lockout\tLock out callout port across different process groups\n");
    fprintf(stderr, "\t^ callout_nohup\tDon't hangup the tty when carrier detect drops\n");
    fprintf(stderr, "\t\t\t\t on the callout device\n");
    fprintf(stderr, "\t^ split_termios Use separate termios for callout and dailin lines\n");
    fprintf(stderr, "\t^ hup_notify\tNotify a process blocked on opening a dial in line\n");
    fprintf(stderr, "\t\t\t\twhen a process has finished using a callout\n");
    fprintf(stderr, "\t\t\t\tline by returning EAGAIN to the open.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t  spd_hi\tuse 56kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_vhi\tuse 115kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_shi\tuse 230kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp\tuse 460kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp13\tuse 920kb instead of 38.4kb (RocketPort Plus cards only!)\n");
    fprintf(stderr, "\t  spd_normal\tuse 38.4kb when a baud rate of 38.4kb is selected\n");
    fprintf(stderr, "\t* country\tspecify RocketModem i country code\n");
    fprintf(stderr, "\n");
    exit(1);
}

main(int argc, char **argv)
{
    int	get_flag = 0;
    int	mkdev_flag = 0;
    int	c;
    extern int optind;
    extern char *optarg;
	
    progname = argv[0];
    if (argc == 1)
	usage();
    while ((c = getopt(argc, argv, "abgoqvMVW")) != EOF) {
	switch (c) {
	case 'a':
	    verbosity = 2;
	    break;
	case 'b':
	    verbosity = 0;
	    break;
	case 'q':
	    quiet_flag++;
	    break;
	case 'v':
	    verbose_flag++;
	    break;
	case 'g':
	    get_flag++;
	    mkdev_flag = 0;
	    break;
	case 'o':
	    old_device_name++;
	    break;
	case 'M':
	    mkdev_flag++;
	    get_flag = 0;
	    break;
	case 'V':
	    fprintf(stderr, "setrocket version %s\n", VERSION_STR);
	    exit(0);
	default:
	    usage();
	}
    }
    if (get_flag) {
	argv += optind;
	while (*argv)
	    get_rocket(*argv++);
	exit(0);
    }
    if (argc == optind)
	usage();
    if (mkdev_flag) {
	mkdev_rocket(argv[optind]);
	exit(0);
    }
    if (argc-optind == 1)
	get_rocket(argv[optind]);
    else
	set_rocket(argv[optind], argv+optind+1);
    exit(0);
}

